@@ -17,7 +17,7 @@
   +----------------------------------------------------------------------+
 */
 
-/* $Id: memcache.c,v 1.80 2007/03/06 18:27:18 tony2001 Exp $ */
+/* $Id: memcache.c,v 1.86 2007/09/02 10:48:13 mikl Exp $ */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -35,11 +35,11 @@
 
 #include <zlib.h>
 #include <time.h>
+#include "ext/standard/crc32.h"
 #include "ext/standard/info.h"
 #include "ext/standard/php_string.h"
 #include "ext/standard/php_var.h"
 #include "ext/standard/php_smart_str.h"
-#include "ext/standard/crc32.h"
 #include "php_network.h"
 #include "php_memcache.h"
 
@@ -154,35 +154,54 @@ static PHP_INI_MH(OnUpdateFailoverAttempts) /* {{{ */
 }
 /* }}} */
 
+static PHP_INI_MH(OnUpdateHashStrategy) /* {{{ */
+{
+	if (!strcasecmp(new_value, "standard")) {
+		MEMCACHE_G(hash_strategy) = MMC_STANDARD_HASH;
+	}
+	else if (!strcasecmp(new_value, "consistent")) {
+		MEMCACHE_G(hash_strategy) = MMC_CONSISTENT_HASH;
+	}
+	else {
+		php_error_docref(NULL TSRMLS_CC, E_WARNING, "memcache.hash_strategy must be in set {standard, consistent} ('%s' given)", new_value);
+		return FAILURE;
+	}
+
+	return SUCCESS;
+}
+/* }}} */
+
+static PHP_INI_MH(OnUpdateHashFunction) /* {{{ */
+{
+	if (!strcasecmp(new_value, "crc32")) {
+		MEMCACHE_G(hash_function) = MMC_HASH_CRC32;
+	}
+	else if (!strcasecmp(new_value, "fnv")) {
+		MEMCACHE_G(hash_function) = MMC_HASH_FNV1A;
+	}
+	else {
+		php_error_docref(NULL TSRMLS_CC, E_WARNING, "memcache.hash_function must be in set {crc32, fnv} ('%s' given)", new_value);
+		return FAILURE;
+	}
+
+	return SUCCESS;
+}
+/* }}} */
+
 /* {{{ PHP_INI */
 PHP_INI_BEGIN()
 	STD_PHP_INI_ENTRY("memcache.allow_failover",	"1",		PHP_INI_ALL, OnUpdateLong,		allow_failover,	zend_memcache_globals,	memcache_globals)
 	STD_PHP_INI_ENTRY("memcache.max_failover_attempts",	"20",	PHP_INI_ALL, OnUpdateFailoverAttempts,		max_failover_attempts,	zend_memcache_globals,	memcache_globals)
 	STD_PHP_INI_ENTRY("memcache.default_port",		"11211",	PHP_INI_ALL, OnUpdateLong,		default_port,	zend_memcache_globals,	memcache_globals)
 	STD_PHP_INI_ENTRY("memcache.chunk_size",		"8192",		PHP_INI_ALL, OnUpdateChunkSize,	chunk_size,		zend_memcache_globals,	memcache_globals)
+	STD_PHP_INI_ENTRY("memcache.hash_strategy",		"standard",	PHP_INI_ALL, OnUpdateHashStrategy,	hash_strategy,	zend_memcache_globals,	memcache_globals)
+	STD_PHP_INI_ENTRY("memcache.hash_function",		"crc32",	PHP_INI_ALL, OnUpdateHashFunction,	hash_function,	zend_memcache_globals,	memcache_globals)
 PHP_INI_END()
 /* }}} */
 
 /* {{{ macros */
 #define MMC_PREPARE_KEY(key, key_len) \
 	php_strtr(key, key_len, "\t\r\n ", "____", 4); \
-
-#if ZEND_DEBUG
-
-#define MMC_DEBUG(info) \
-{\
-	mmc_debug info; \
-}\
-
-#else
-
-#define MMC_DEBUG(info) \
-{\
-}\
-
-#endif
-
-
 /* }}} */
 
 /* {{{ internal function protos */
@@ -194,11 +213,9 @@ static void mmc_server_disconnect(mmc_t * TSRMLS_DC);
 static void mmc_server_deactivate(mmc_t * TSRMLS_DC);
 static int mmc_server_store(mmc_t *, const char *, int TSRMLS_DC);
 
-static unsigned int mmc_hash(const char *, int);
 static int mmc_compress(char **, unsigned long *, const char *, int TSRMLS_DC);
 static int mmc_uncompress(char **, unsigned long *, const char *, int);
 static int mmc_get_pool(zval *, mmc_pool_t ** TSRMLS_DC);
-static int mmc_open(mmc_t *, int, char **, int * TSRMLS_DC);
 static int mmc_close(mmc_t * TSRMLS_DC);
 static int mmc_readline(mmc_t * TSRMLS_DC);
 static char * mmc_get_version(mmc_t * TSRMLS_DC);
@@ -215,6 +232,11 @@ static void php_mmc_incr_decr(INTERNAL_FUNCTION_PARAMETERS, int);
 static void php_mmc_connect(INTERNAL_FUNCTION_PARAMETERS, int);
 /* }}} */
 
+/* {{{ hash strategies */
+extern mmc_hash_t mmc_standard_hash;
+extern mmc_hash_t mmc_consistent_hash;
+/* }}} */
+
 /* {{{ php_memcache_init_globals()
 */
 static void php_memcache_init_globals(zend_memcache_globals *memcache_globals_p TSRMLS_DC)
@@ -222,6 +244,8 @@ static void php_memcache_init_globals(zend_memcache_globals *memcache_globals_p
 	MEMCACHE_G(debug_mode)		  = 0;
 	MEMCACHE_G(num_persistent)	  = 0;
 	MEMCACHE_G(compression_level) = Z_DEFAULT_COMPRESSION;
+	MEMCACHE_G(hash_strategy)	  = MMC_STANDARD_HASH;
+	MEMCACHE_G(hash_function)	  = MMC_HASH_CRC32;
 }
 /* }}} */
 
@@ -285,7 +309,7 @@ PHP_MINFO_FUNCTION(memcache)
 	php_info_print_table_start();
 	php_info_print_table_header(2, "memcache support", "enabled");
 	php_info_print_table_row(2, "Active persistent connections", buf);
-	php_info_print_table_row(2, "Revision", "$Revision: 1.80 $");
+	php_info_print_table_row(2, "Revision", "$Revision: 1.86 $");
 	php_info_print_table_end();
 
 	DISPLAY_INI_ENTRIES();
@@ -297,7 +321,7 @@ PHP_MINFO_FUNCTION(memcache)
    ------------------ */
 
 #if ZEND_DEBUG
-static void mmc_debug(const char *format, ...) /* {{{ */
+void mmc_debug(const char *format, ...) /* {{{ */
 {
 	TSRMLS_FETCH();
 
@@ -454,7 +478,8 @@ static int mmc_server_store(mmc_t *mmc, const char *request, int request_len TSR
 	}
 
 	/* return FALSE without failover */
-	if (mmc_str_left(mmc->inbuf, "SERVER_ERROR out of memory", response_len, sizeof("SERVER_ERROR out of memory") - 1)) {
+	if (mmc_str_left(mmc->inbuf, "SERVER_ERROR out of memory", response_len, sizeof("SERVER_ERROR out of memory") - 1) ||
+		mmc_str_left(mmc->inbuf, "SERVER_ERROR object too large", response_len, sizeof("SERVER_ERROR object too large")-1)) {
 		return 0;
 	}
 
@@ -462,14 +487,62 @@ static int mmc_server_store(mmc_t *mmc, const char *request, int request_len TSR
 }
 /* }}} */
 
-mmc_pool_t *mmc_pool_new() /* {{{ */
+static unsigned int mmc_hash_crc32(const char *key, int key_len) /* 
+	CRC32 hash {{{ */
 {
+	unsigned int crc = ~0;
+	int i;
+
+	for (i=0; i<key_len; i++) {
+		CRC32(crc, key[i]);
+	}
+
+  	return ~crc;
+}
+/* }}} */
+
+static unsigned int mmc_hash_fnv1a(const char *key, int key_len) /* 
+	FNV-1a hash {{{ */
+{
+	unsigned int hval = FNV_32_INIT;
+	int i;
+
+	for (i=0; i<key_len; i++) {
+		hval ^= (unsigned int)key[i];
+		hval *= FNV_32_PRIME;
+    }
+
+    return hval;
+}
+/* }}} */
+
+mmc_pool_t *mmc_pool_new(TSRMLS_D) /* {{{ */
+{
+	mmc_hash_function hash;
 	mmc_pool_t *pool = emalloc(sizeof(mmc_pool_t));
 	pool->num_servers = 0;
-	pool->num_buckets = 0;
 	pool->compress_threshold = 0;
 	pool->in_free = 0;
 	pool->min_compress_savings = MMC_DEFAULT_SAVINGS;
+
+	switch (MEMCACHE_G(hash_strategy)) {
+		case MMC_CONSISTENT_HASH:
+			pool->hash = &mmc_consistent_hash;
+			break;
+		default:
+			pool->hash = &mmc_standard_hash;
+	}
+	
+	switch (MEMCACHE_G(hash_function)) {
+		case MMC_HASH_FNV1A:
+			hash = &mmc_hash_fnv1a;
+			break;
+		default:
+			hash = &mmc_hash_crc32;
+	}
+	
+	pool->hash_state = pool->hash->create_state(hash);
+
 	return pool;
 }
 /* }}} */
@@ -502,18 +575,13 @@ void mmc_pool_free(mmc_pool_t *pool TSRMLS_DC) /* {{{ */
 		efree(pool->requests);
 	}
 
-	if (pool->num_buckets) {
-		efree(pool->buckets);
-	}
-
+	pool->hash->free_state(pool->hash_state);
 	efree(pool);
 }
 /* }}} */
 
 void mmc_pool_add(mmc_pool_t *pool, mmc_t *mmc, unsigned int weight) /* {{{ */
 {
-	int i;
-
 	/* add server and a preallocated request pointer */
 	if (pool->num_servers) {
 		pool->servers = erealloc(pool->servers, sizeof(mmc_t *) * (pool->num_servers + 1));
@@ -527,47 +595,7 @@ void mmc_pool_add(mmc_pool_t *pool, mmc_t *mmc, unsigned int weight) /* {{{ */
 	pool->servers[pool->num_servers] = mmc;
 	pool->num_servers++;
 
-	/* add weight number of buckets for this server */
-	if (pool->num_buckets) {
-		pool->buckets = erealloc(pool->buckets, sizeof(mmc_t *) * (pool->num_buckets + weight));
-	}
-	else {
-		pool->buckets = emalloc(sizeof(mmc_t *) * (pool->num_buckets + weight));
-	}
-
-	for (i=0; i<weight; i++) {
-		pool->buckets[pool->num_buckets + i] = mmc;
-	}
-	pool->num_buckets += weight;
-}
-/* }}} */
-
-mmc_t *mmc_pool_find(mmc_pool_t *pool, const char *key, int key_len TSRMLS_DC) /* {{{ */
-{
-	mmc_t *mmc;
-
-	if (pool->num_servers > 1) {
-		unsigned int hash = mmc_hash(key, key_len), i;
-		mmc = pool->buckets[hash % pool->num_buckets];
-
-		/* perform failover if needed */
-		for (i=0; !mmc_open(mmc, 0, NULL, NULL TSRMLS_CC) && MEMCACHE_G(allow_failover) && i<MEMCACHE_G(max_failover_attempts); i++) {
-			char *next_key = emalloc(key_len + MAX_LENGTH_OF_LONG + 1);
-			int next_len = sprintf(next_key, "%d%s", i+1, key);
-			MMC_DEBUG(("mmc_pool_find: failed to connect to server '%s:%d' status %d, trying next", mmc->host, mmc->port, mmc->status));
-
-			hash += mmc_hash(next_key, next_len);
-			mmc = pool->buckets[hash % pool->num_buckets];
-
-			efree(next_key);
-		}
-	}
-	else {
-		mmc = pool->servers[0];
-		mmc_open(mmc, 0, NULL, NULL TSRMLS_CC);
-	}
-
-	return mmc->status != MMC_STATUS_FAILED ? mmc : NULL;
+	pool->hash->add_server(pool->hash_state, mmc, weight);
 }
 /* }}} */
 
@@ -720,22 +748,6 @@ static int mmc_uncompress(char **result, unsigned long *result_len, const char *
 }
 /* }}}*/
 
-/**
- * New style crc32 hash, compatible with other clients
- */
-static unsigned int mmc_hash(const char *key, int key_len) { /* {{{ */
-	unsigned int crc = ~0;
-	int i;
-
-	for (i=0; i<key_len; i++) {
-		CRC32(crc, key[i]);
-	}
-
-	crc = (~crc >> 16) & 0x7fff;
-  	return crc ? crc : 1;
-}
-/* }}} */
-
 static int mmc_get_pool(zval *id, mmc_pool_t **pool TSRMLS_DC) /* {{{ */
 {
 	zval **connection;
@@ -842,7 +854,7 @@ static int _mmc_open(mmc_t *mmc, char **error_string, int *errnum TSRMLS_DC) /*
 }
 /* }}} */
 
-static int mmc_open(mmc_t *mmc, int force_connect, char **error_string, int *errnum TSRMLS_DC) /* {{{ */
+int mmc_open(mmc_t *mmc, int force_connect, char **error_string, int *errnum TSRMLS_DC) /* {{{ */
 {
 	switch (mmc->status) {
 		case MMC_STATUS_DISCONNECTED:
@@ -942,7 +954,7 @@ static int mmc_readline(mmc_t *mmc TSRMLS_DC) /* {{{ */
 		return -1;
 	}
 
-	response = php_stream_get_line(mmc->stream, mmc->inbuf, MMC_BUF_SIZE, &response_len);
+	response = php_stream_get_line(mmc->stream, ZSTR(mmc->inbuf), MMC_BUF_SIZE, &response_len);
 	if (response) {
 		MMC_DEBUG(("mmc_readline: read data:"));
 		MMC_DEBUG(("mmc_readline:---"));
@@ -1140,7 +1152,7 @@ static int mmc_exec_retrieval_cmd_multi(mmc_pool_t *pool, zval *keys, zval **ret
 				FREE_ZVAL(tmp);
 			}
 
-			MMC_PREPARE_KEY(str_key, Z_STRLEN_PP(key));
+			MMC_PREPARE_KEY(str_key, key_len);
 
 			/* schedule key if first round or if missing from result */
 			if ((!i || !zend_hash_exists(Z_ARRVAL_PP(return_value), str_key, key_len)) &&
@@ -1632,19 +1644,30 @@ static void php_mmc_store(INTERNAL_FUNCTION_PARAMETERS, char *command, int comma
 			value_len = Z_STRLEN_P(var);
 			break;
 
-		default:
+		default: {
+			zval var_copy, *var_copy_ptr;
+
+			/* FIXME: we should be using 'Z' instead of this, but unfortunately it's PHP5-only */
+			var_copy = *var;
+			zval_copy_ctor(&var_copy);
+
+			var_copy_ptr = &var_copy;
+
 			PHP_VAR_SERIALIZE_INIT(var_hash);
-			php_var_serialize(&buf, &var, &var_hash TSRMLS_CC);
+			php_var_serialize(&buf, &var_copy_ptr, &var_hash TSRMLS_CC);
 			PHP_VAR_SERIALIZE_DESTROY(var_hash);
 
 			if (!buf.c) {
-				/* you're trying to save null or something went really wrong */
+				/* something went really wrong */
+				zval_dtor(&var_copy);
 				RETURN_FALSE;
 			}
 
 			value = buf.c;
 			value_len = buf.len;
 			flags |= MMC_SERIALIZED;
+			zval_dtor(&var_copy);
+		}
 			break;
 	}
 
@@ -1740,7 +1763,7 @@ static void php_mmc_connect (INTERNAL_FUNCTION_PARAMETERS, int persistent) /* {{
 
 	/* initialize pool and object if need be */
 	if (!mmc_object) {
-		pool = mmc_pool_new();
+		pool = mmc_pool_new(TSRMLS_C);
 		mmc_pool_add(pool, mmc, 1);
 
 		object_init_ex(return_value, memcache_class_entry_ptr);
@@ -1758,7 +1781,7 @@ static void php_mmc_connect (INTERNAL_FUNCTION_PARAMETERS, int persistent) /* {{
 		RETURN_TRUE;
 	}
 	else {
-		pool = mmc_pool_new();
+		pool = mmc_pool_new(TSRMLS_C);
 		mmc_pool_add(pool, mmc, 1);
 
 		list_id = zend_list_insert(pool, le_memcache_pool);
@@ -1798,7 +1821,7 @@ PHP_FUNCTION(memcache_add_server)
 	long port = MEMCACHE_G(default_port), weight = 1, timeout = MMC_DEFAULT_TIMEOUT, retry_interval = MMC_DEFAULT_RETRY;
 	zend_bool persistent = 1, status = 1;
 	int resource_type, host_len, list_id;
-	char *host, *failure_callback_name;
+	char *host;
 
 	if (mmc_object) {
 		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|lblllbz", &host, &host_len, &port, &persistent, &weight, &timeout, &retry_interval, &status, &failure_callback) == FAILURE) {
@@ -1817,12 +1840,10 @@ PHP_FUNCTION(memcache_add_server)
 	}
 
 	if (failure_callback != NULL && Z_TYPE_P(failure_callback) != IS_NULL) {
-		if (!zend_is_callable(failure_callback, 0, &failure_callback_name)) {
-			php_error_docref(NULL TSRMLS_CC, E_WARNING, "invalid failure callback '%s' passed", failure_callback_name);
-			efree(failure_callback_name);
+		if (!zend_is_callable(failure_callback, 0, NULL)) {
+			php_error_docref(NULL TSRMLS_CC, E_WARNING, "invalid failure callback passed");
 			RETURN_FALSE;
 		}
-		efree(failure_callback_name);
 	}
 
 	/* lazy initialization of server struct */
@@ -1846,7 +1867,7 @@ PHP_FUNCTION(memcache_add_server)
 
 	/* initialize pool if need be */
 	if (zend_hash_find(Z_OBJPROP_P(mmc_object), "connection", sizeof("connection"), (void **) &connection) == FAILURE) {
-		pool = mmc_pool_new();
+		pool = mmc_pool_new(TSRMLS_C);
 		list_id = zend_list_insert(pool, le_memcache_pool);
 		add_property_resource(mmc_object, "connection", list_id);
 	}
@@ -1873,7 +1894,7 @@ PHP_FUNCTION(memcache_set_server_params)
 	long port = MEMCACHE_G(default_port), timeout = MMC_DEFAULT_TIMEOUT, retry_interval = MMC_DEFAULT_RETRY;
 	zend_bool status = 1;
 	int host_len, i;
-	char *host, *failure_callback_name;
+	char *host;
 
 	if (mmc_object) {
 		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|lllbz", &host, &host_len, &port, &timeout, &retry_interval, &status, &failure_callback) == FAILURE) {
@@ -1903,12 +1924,10 @@ PHP_FUNCTION(memcache_set_server_params)
 	}
 
 	if (failure_callback != NULL && Z_TYPE_P(failure_callback) != IS_NULL) {
-		if (!zend_is_callable(failure_callback, 0, &failure_callback_name)) {
-			php_error_docref(NULL TSRMLS_CC, E_WARNING, "invalid failure callback '%s' passed", failure_callback_name);
-			efree(failure_callback_name);
+		if (!zend_is_callable(failure_callback, 0, NULL)) {
+			php_error_docref(NULL TSRMLS_CC, E_WARNING, "invalid failure callback passed");
 			RETURN_FALSE;
 		}
-		efree(failure_callback_name);
 	}
 
 	mmc->timeout = timeout;
@@ -1984,7 +2003,7 @@ PHP_FUNCTION(memcache_get_server_status)
 mmc_t *mmc_find_persistent(char *host, int host_len, int port, int timeout, int retry_interval TSRMLS_DC) /* {{{ */
 {
 	mmc_t *mmc;
-	list_entry *le;
+	zend_rsrc_list_entry *le;
 	char *hash_key;
 	int hash_key_len;
 
@@ -1993,7 +2012,7 @@ mmc_t *mmc_find_persistent(char *host, int host_len, int port, int timeout, int
 	hash_key_len = sprintf(hash_key, "mmc_connect___%s:%d", host, port);
 
 	if (zend_hash_find(&EG(persistent_list), hash_key, hash_key_len+1, (void **) &le) == FAILURE) {
-		list_entry new_le;
+		zend_rsrc_list_entry new_le;
 		MMC_DEBUG(("mmc_find_persistent: connection wasn't found in the hash"));
 
 		mmc = mmc_server_new(host, host_len, port, 1, timeout, retry_interval TSRMLS_CC);
@@ -2001,7 +2020,7 @@ mmc_t *mmc_find_persistent(char *host, int host_len, int port, int timeout, int
 		new_le.ptr  = mmc;
 
 		/* register new persistent connection */
-		if (zend_hash_update(&EG(persistent_list), hash_key, hash_key_len+1, (void *) &new_le, sizeof(list_entry), NULL) == FAILURE) {
+		if (zend_hash_update(&EG(persistent_list), hash_key, hash_key_len+1, (void *) &new_le, sizeof(zend_rsrc_list_entry), NULL) == FAILURE) {
 			mmc_server_free(mmc TSRMLS_CC);
 			mmc = NULL;
 		} else {
@@ -2009,7 +2028,7 @@ mmc_t *mmc_find_persistent(char *host, int host_len, int port, int timeout, int
 		}
 	}
 	else if (le->type != le_pmemcache || le->ptr == NULL) {
-		list_entry new_le;
+		zend_rsrc_list_entry new_le;
 		MMC_DEBUG(("mmc_find_persistent: something was wrong, reconnecting.."));
 		zend_hash_del(&EG(persistent_list), hash_key, hash_key_len+1);
 
@@ -2018,7 +2037,7 @@ mmc_t *mmc_find_persistent(char *host, int host_len, int port, int timeout, int
 		new_le.ptr  = mmc;
 
 		/* register new persistent connection */
-		if (zend_hash_update(&EG(persistent_list), hash_key, hash_key_len+1, (void *) &new_le, sizeof(list_entry), NULL) == FAILURE) {
+		if (zend_hash_update(&EG(persistent_list), hash_key, hash_key_len+1, (void *) &new_le, sizeof(zend_rsrc_list_entry), NULL) == FAILURE) {
 			mmc_server_free(mmc TSRMLS_CC);
 			mmc = NULL;
 		}
